import json
from typing import TYPE_CHECKING

from django.http import HttpRequest, HttpResponseBadRequest
from graphene_django.views import GraphQLView, HttpError

from .files import place_files
from .typing import Any

if TYPE_CHECKING:
    from django.core.files import File

__all__ = [
    "FileUploadGraphQLView",
]


class FileUploadGraphQLView(GraphQLView):
    def parse_body(self, request: HttpRequest) -> dict[str, Any]:
        if self.is_file_upload(request):
            return self.parse_file_uploads(request)
        return super().parse_body(request)

    def is_file_upload(self, request: HttpRequest) -> bool:
        content_type = self.get_content_type(request)
        return content_type == "multipart/form-data" and request.FILES

    def parse_file_uploads(self, request: HttpRequest) -> dict[str, Any]:
        operations = self.get_operations(request)
        files_map = self.get_map(request)
        files: dict[str, File] = request.FILES  # type: ignore[assignment]
        place_files(operations, files_map, files)
        return operations

    @staticmethod
    def get_operations(request: HttpRequest) -> dict[str, Any]:  # pragma: no cover
        operations_str: str | None = request.POST.get("operations")
        if not isinstance(operations_str, str):
            msg = "File upload must contain an `operations` value."
            raise HttpError(HttpResponseBadRequest(msg))

        try:
            operations: dict[str, Any] = json.loads(operations_str)
        except json.JSONDecodeError as error:
            msg = "The `operations` value must be a JSON object."
            raise HttpError(HttpResponseBadRequest(msg)) from error

        if not isinstance(operations, dict):
            msg = "The `operations` value is not a mapping."
            raise HttpError(HttpResponseBadRequest(msg))

        return operations

    @staticmethod
    def get_map(request: HttpRequest) -> dict[str, list[str]]:  # pragma: no cover
        files_map_str: str | None = request.POST.get("map")
        if not isinstance(files_map_str, str):
            msg = "File upload must contain an `map` value."
            raise HttpError(HttpResponseBadRequest(msg))

        try:
            files_map: dict[str, list[str]] = json.loads(files_map_str)
        except json.JSONDecodeError as error:
            msg = "The `map` value must be a JSON object."
            raise HttpError(HttpResponseBadRequest(msg)) from error

        if not isinstance(files_map, dict):
            msg = "The `map` value is not a mapping."
            raise HttpError(HttpResponseBadRequest(msg))

        for value in files_map.values():
            if not isinstance(value, list) or not all(isinstance(item, str) for item in value):
                msg = "The `map` value is not a mapping from string to list of strings."
                raise HttpError(HttpResponseBadRequest(msg))

        return files_map



from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
from functools import partial
from inspect import cleandoc
from io import BytesIO
from typing import Any, Callable, Generator, NamedTuple, TypedDict, TypeVar

import pytest
import sqlparse
from django import db
from django.core.files import File

__all__ = [
    "capture_database_queries",
    "compare_unordered",
    "create_mock_png",
    "parametrize_helper",
]


@dataclass
class QueryData:
    queries: list[str]

    @property
    def log(self) -> str:
        message = "-" * 75
        message += f"\n>>> Queries ({len(self.queries)}):\n"
        for index, query in enumerate(self.queries):
            formatted_query = sqlparse.format(query, reindent=True)
            message += f"{index + 1}) ".ljust(75, "-") + f"\n{formatted_query}\n"
        message += "-" * 75
        return message


def _db_query_logger(  # noqa: PLR0913
    execute: Callable[..., Any],
    sql: str,
    params: tuple[Any, ...],
    many: bool,  # noqa: FBT001
    context: dict[str, Any],
    # Added with functools.partial()
    query_cache: list[str],
) -> Any:
    """
    A database query logger for capturing executed database queries.
    Used to check that query optimizations work as expected.

    Can also be used as a place to put debugger breakpoint for solving issues.
    """
    # Don't include transaction creation, as we aren't interested in them.
    if not sql.startswith("SAVEPOINT") and not sql.startswith("RELEASE SAVEPOINT"):
        try:
            query_cache.append(sql % params)
        except TypeError:  # pragma: no cover
            query_cache.append(sql)
    return execute(sql, params, many, context)


@contextmanager
def capture_database_queries() -> Generator[QueryData, None, None]:
    """Capture results of what database queries were executed. `DEBUG` needs to be set to True."""
    results = QueryData(queries=[])
    query_logger = partial(_db_query_logger, query_cache=results.queries)

    with db.connection.execute_wrapper(query_logger):
        yield results


TNamedTuple = TypeVar("TNamedTuple", bound=NamedTuple)


class ParametrizeArgs(TypedDict):
    argnames: list[str]
    argvalues: list[TNamedTuple]
    ids: list[str]


def parametrize_helper(__tests: dict[str, TNamedTuple], /) -> ParametrizeArgs:
    """Construct parametrize input while setting test IDs."""
    assert __tests, "I need some tests, please!"  # noqa: S101
    values = list(__tests.values())
    try:
        return ParametrizeArgs(
            argnames=list(values[0].__class__.__annotations__),
            argvalues=values,
            ids=list(__tests),
        )
    except AttributeError as error:  # pragma: no cover
        msg = "Improper configuration. Did you use a NamedTuple for TNamedTuple?"
        raise RuntimeError(msg) from error


def _dict_sorter(data: dict[str, Any]) -> tuple[str, str]:
    return str(data.keys()), str(data.values())


class ComparisonError(Exception):
    errors_by_code: dict[int, str] = {
        1: "type mismatch",
        2: "value mismatch",
        3: "dict keys mismatch",
        4: "array length mismatch",
        5: "array content mismatch",
        6: "array type mismatch",
    }

    def __init__(self, actual: Any, expected: Any, *, reason_code: int) -> None:
        self.actual = actual
        self.expected = expected
        self.actual_key = ""
        self.expected_key = ""
        self.reason = self.errors_by_code[reason_code]


def compare_unordered(actual: dict[str, Any] | list[Any], expected: dict[str, Any] | list[Any]) -> None:
    """
    Compare two json data structures without caring about the order in which items appear in lists.

    If the comparison fails, will raise a `pytest.fail` with an error message
    which will try to point out the place where the comparison failed.
    However, this will not always pick the best "nodes" to show, so
    full comparison data will also be shown in the error message for clarity.

    Disclaimer: Cannot compare json with heterogeneous lists.
    """
    msg: str = ""
    try:
        _compare_unordered(actual, expected)
    except ComparisonError as error:
        msg = cleandoc(
            f"""
            Actual and expected data don't match ({error.reason}):
                actual{error.actual_key} = {error.actual}
                expected{error.expected_key} = {error.expected}
            """
        )
        if error.actual_key or error.expected_key:
            msg = (
                msg
                + "\n"
                + cleandoc(
                    f"""
                    Full data:
                        actual = {actual}
                        expected = {expected}
                    """
                )
            )
    finally:
        if msg:
            pytest.fail(msg, pytrace=False)


def _compare_unordered(actual: Any, expected: Any) -> None:  # noqa: PLR0912, C901
    if type(actual) is not type(expected):
        raise ComparisonError(actual, expected, reason_code=1)

    if isinstance(actual, dict):
        if actual.keys() != expected.keys():
            raise ComparisonError(actual, expected, reason_code=3)

        for key in actual:
            try:
                _compare_unordered(actual[key], expected[key])
            except ComparisonError as error:  # noqa: PERF203
                error.actual_key = f"['{key}']{error.actual_key}"
                error.expected_key = f"['{key}']{error.expected_key}"
                raise

    elif isinstance(actual, list):
        if len(actual) != len(expected):
            raise ComparisonError(actual, expected, reason_code=4)

        actual_types = {type(item) for item in actual}
        expected_types = {type(item) for item in expected}

        if len(actual_types) != 1 or len(expected_types) != 1:
            raise ComparisonError(actual, expected, reason_code=6)

        actual_type = actual_types.pop()
        expected_type = expected_types.pop()
        if actual_type is not expected_type:
            raise ComparisonError(actual, expected, reason_code=5)

        if issubclass(actual_type, dict):
            for act, exp in zip(sorted(actual, key=_dict_sorter), sorted(expected, key=_dict_sorter)):
                try:
                    _compare_unordered(act, exp)
                except ComparisonError as error:  # noqa: PERF203
                    error.actual_key = f"[-]{error.actual_key}"
                    error.expected_key = f"[-]{error.expected_key}"
                    raise
        else:
            for act, exp in zip(sorted(actual), sorted(expected)):
                try:
                    _compare_unordered(act, exp)
                except ComparisonError as error:  # noqa: PERF203
                    error.actual_key = f"[-]{error.actual_key}"
                    error.expected_key = f"[-]{error.expected_key}"
                    raise

    elif actual != expected:
        raise ComparisonError(actual, expected, reason_code=2)


def create_mock_png() -> File:
    from PIL import Image

    bytes_io = BytesIO()
    img = Image.new("RGB", (1, 1))
    img.save(bytes_io, format="PNG", compress_level=1)
    bytes_io.seek(0)
    return File(bytes_io, name="image.png")



from __future__ import annotations

import datetime
from contextlib import suppress
from typing import TYPE_CHECKING

import graphene
import graphql
from aniso8601 import parse_time
from django.db import models
from graphene.types.generic import GenericScalar
from graphene.types.inputobjecttype import InputObjectTypeOptions
from graphene.types.utils import get_field_as
from graphene_django.converter import convert_django_field
from graphql import Undefined
from query_optimizer import DjangoConnectionField, DjangoListField, RelatedField

from ..typing import StrEnum
from ..utils import get_operator_enum

if TYPE_CHECKING:
    from django.db.models import Model
    from graphene.types.unmountedtype import UnmountedType

    from graphene_django_extensions.typing import Any, FieldAliasToLookup, Self, TypedDict


__all__ = [
    "DjangoConnectionField",
    "DjangoListField",
    "Duration",
    "OrderingChoices",
    "RelatedField",
    "Time",
    "TypedDictField",
    "TypedDictListField",
    "Upload",
    "UserDefinedFilterInputType",
]


class Time(graphene.Time):
    """Time scalar that can parse time-strings from database."""

    @staticmethod
    def serialize(time: datetime.time | str) -> str:
        if isinstance(time, str):
            with suppress(ValueError):
                time = parse_time(time)
        return graphene.Time.serialize(time)


class Duration(graphene.Scalar):
    """Represents a DurationField value as an integer in seconds."""

    @staticmethod
    def serialize(value: Any) -> int | None:
        if not isinstance(value, datetime.timedelta):  # pragma: no cover
            return None
        return int(value.total_seconds())

    @staticmethod
    def parse_value(value: Any) -> datetime.timedelta | None:
        if not isinstance(value, int):  # pragma: no cover
            return None
        return datetime.timedelta(seconds=value)

    @staticmethod
    def parse_literal(ast: graphql.ValueNode, _variables: Any = None) -> int | Undefined:  # pragma: no cover
        if isinstance(ast, graphql.IntValueNode):
            return int(ast.value)
        return Undefined


class Upload(graphene.Scalar):  # pragma: no cover
    """Represents a file upload."""

    @staticmethod
    def serialize(value: Any) -> Any:
        return value

    @staticmethod
    def parse_literal(node: Any) -> Any:
        return node

    @staticmethod
    def parse_value(value: Any) -> Any:
        return value


class TypedDictFieldMixin:
    CONVERSION_TABLE: dict[type, type[models.Field]] = {
        int: models.IntegerField,
        str: models.CharField,
        bool: models.BooleanField,
        float: models.FloatField,
        dict: models.JSONField,
        list: models.JSONField,
        set: models.JSONField,
        tuple: models.JSONField,
        bytes: models.BinaryField,
        datetime.datetime: models.DateTimeField,
        datetime.date: models.DateField,
        datetime.time: models.TimeField,
    }

    def __init__(self, typed_dict: type[TypedDict], *arg: Any, **kwargs: Any) -> None:  # pragma: no cover
        type_ = self.convert_typed_dict_to_graphene_type(typed_dict)
        super().__init__(type_, *arg, **kwargs)

    def convert_typed_dict_to_graphene_type(self, typed_dict: type[TypedDict]) -> type[graphene.ObjectType]:
        graphene_types: dict[str, UnmountedType] = {}
        for field_name, type_ in typed_dict.__annotations__.items():
            model_field = self.CONVERSION_TABLE.get(type_)
            if model_field is None:
                msg = f"Cannot convert field `{field_name}` of type `{type_.__name__}` to model field."
                raise ValueError(msg)
            graphene_type = convert_django_field(model_field())
            graphene_types[field_name] = graphene_type

        return type(f"{typed_dict.__name__}Type", (graphene.ObjectType,), graphene_types)  # type: ignore[return-value]


class TypedDictField(TypedDictFieldMixin, graphene.Field):
    """Field that converts a TypedDict to a graphene Field ObjectType."""


class TypedDictListField(TypedDictFieldMixin, graphene.List):
    """Field that converts a TypedDict to a graphene List ObjectType."""


class OrderingChoices(graphene.Enum):
    """Get `order_by` choices with Enums rather than strings."""

    @classmethod
    def create(cls, model: type[Model], fields_map: dict[str, str]) -> type[Self]:
        name = f"{model.__name__}OrderingChoices"
        meta_dict = {"model": model, "fields_map": fields_map}
        attrs = {"Meta": type("Meta", (), meta_dict)}
        return type(name, (OrderingChoices,), attrs)

    @classmethod
    def __init_subclass_with_meta__(
        cls,
        model: type[Model] | None = None,
        fields_map: dict[str, str] | None = None,
        **options: Any,
    ) -> None:
        if model is None:  # pragma: no cover
            msg = "'Meta.model' is required."
            raise TypeError(msg)

        if fields_map is None:  # pragma: no cover
            msg = "'Meta.fields' is required."
            raise TypeError(msg)

        enum = StrEnum(f"{model.__name__}OrderingChoices", fields_map)
        description = f"Ordering fields for the '{model.__name__}' model."

        super().__init_subclass_with_meta__(enum=enum, description=description, **options)


class UserDefinedFilterInputType(graphene.InputObjectType):
    """User defined filtering input."""

    operation = graphene.Field(graphene.Enum.from_enum(get_operator_enum()), required=True)
    value = GenericScalar()

    @classmethod
    def create(cls, model: type[Model], fields_map: FieldAliasToLookup) -> type[Self]:
        name = f"{model.__name__}FilterInput"
        meta_dict = {"model": model, "fields_map": fields_map}
        attrs = {"Meta": type("Meta", (), meta_dict)}
        return type(name, (UserDefinedFilterInputType,), attrs)

    @classmethod
    def __init_subclass_with_meta__(
        cls,
        model: type[Model] | None = None,
        fields_map: FieldAliasToLookup | None = None,
        _meta: InputObjectTypeOptions | None = None,
        **options: Any,
    ) -> None:
        if _meta is None:
            _meta = InputObjectTypeOptions(cls)

        if model is None:  # pragma: no cover
            msg = "'Meta.model' is required."
            raise TypeError(msg)

        if fields_map is None:  # pragma: no cover
            msg = "'Meta.fields' is required."
            raise TypeError(msg)

        _meta.fields = {
            "field": graphene.Field(
                graphene.Enum.from_enum(
                    enum=StrEnum(f"{model.__name__}FilterFields", fields_map),
                    description=f"Filterable fields for the '{model.__name__}' model.",
                ),
            ),
            "operations": get_field_as(
                graphene.List(graphene.NonNull(lambda: cls)),
                _as=graphene.InputField,
            ),
        }

        super().__init_subclass_with_meta__(_meta=_meta, **options)



"""
File upload handling utilities.

Compliant with the GraphQL multipart request specification:
https://github.com/jaydenseric/graphql-multipart-request-spec
"""

from __future__ import annotations

from collections import defaultdict
from typing import Any

from django.core.files import File

__all__ = [
    "extract_files",
    "place_files",
]

from django.http import HttpResponseBadRequest
from graphene_django.views import HttpError


def extract_files(variables: dict[str, Any] | list[Any], *, prefix: str = "") -> dict[File, list[str]]:
    """
    Extract Django File objects paths in the given variables. Replace the file objects with None values.

    >>> file_1 = File(name="test_file_1.png")
    >>> file_2 = File(name="test_file_2.png")
    >>>
    >>> variables = {
    ...     "image": file_1,
    ...     "foo": [file_1, file_2],
    ...     "bar": {"one": file_2, "two": file_1},
    ... }
    ...
    >>> files = extract_files(variables, prefix="variables")
    >>> files
    {
        <File: test_file_1.png>: ["variables.image", "variables.foo.0", "variables.bar.two"],
        <File: test_file_2.png>: ["variables.foo.1", "variables.bar.one"],
    }
    >>> variables
    {
        "image": None,
        "foo": [None, None],
        "bar": {"one": None, "two": None}
    }
    """
    prefix = f"{prefix}." if prefix else prefix
    files: dict[File, list[str]] = defaultdict(list)
    iterable = enumerate(variables) if isinstance(variables, list) else variables.items()

    for key, value in iterable:
        if isinstance(value, File):
            variables[key] = None
            files[value].append(f"{prefix}{key}")

        if isinstance(value, (dict, list)):
            for sub_filename, sub_value in extract_files(value, prefix=f"{prefix}{key}").items():
                files[sub_filename] += sub_value

    return files


def place_files(operations: dict[str, Any], files_map: dict[str, list[str]], files: dict[str, File]) -> None:
    """
    Place files in the given operations using the files map.

    >>> file_1 = File(name="test_file_1.png")
    >>> file_2 = File(name="test_file_2.png")
    >>>
    >>> operations = {
    ...     "image": None,
    ...     "foo": [None, None],
    ...     "bar": {"one": None, "two": None},
    ... }
    ...
    >>> files_map = {
    ...     "0": ["image", "foo.0", "bar.two"]
    ...     "1": ["foo.1", "bar.one"],
    ... }
    ...
    >>> files = {
    ...     "0": file_1,
    ...     "1": file_2,
    ... }
    ...
    >>> place_files(operations, files_map, files)
    >>> operations
    {
        "image": file_1,
        "foo": [file_1, file_2],
        "bar": {"one": file_2, "two": file_1},
    }
    """
    for key, values in files_map.items():
        for value in values:
            path: list[str] = value.split(".")
            file: File = files.get(key)
            if file is None:  # pragma: no cover
                msg = f"File for path '{value}' not found in request files."
                raise HttpError(HttpResponseBadRequest(msg))

            _place_file(file, path, operations)


def _place_file(file: File, path: list[str], operations: dict[str, Any] | list[Any]) -> None:
    """Handle placing a single file to a single path in the `operations` object."""
    key: str | int = int(path[0]) if isinstance(operations, list) else path[0]

    try:
        ops: Any = operations[key]
    except (KeyError, IndexError, TypeError) as error:  # pragma: no cover
        msg = "File map does not lead to a null value."
        raise HttpError(HttpResponseBadRequest(msg)) from error

    path_left = path[1:]

    if path_left:
        return _place_file(file, path_left, ops)

    if ops is not None:  # pragma: no cover
        msg = "File map does not lead to a null value."
        raise HttpError(HttpResponseBadRequest(msg))

    operations[key] = file
    return None



from __future__ import annotations

import json
from typing import TYPE_CHECKING

import pytest
from django.contrib.auth import get_user_model
from django.test.client import MULTIPART_CONTENT, Client
from graphene_django.settings import graphene_settings

from ..files import extract_files
from .utils import QueryData, capture_database_queries

if TYPE_CHECKING:
    from django.contrib.auth.models import User
    from django.core.files import File
    from django.http import HttpResponse

    from ..typing import Any, ClassVar, FieldError, Self

__all__ = [
    "GraphQLClient",
]


class GQLResponse:
    def __init__(self, response: HttpResponse, query_data: QueryData) -> None:
        # 'django.test.client.Client.request' sets json attribute on the response.
        self.json: dict[str, Any] = response.json()  # type: ignore[attr-defined]
        self.query_data = query_data

    def __str__(self) -> str:
        return json.dumps(self.json, indent=2, sort_keys=True, default=str)

    def __repr__(self) -> str:
        return repr(self.json)

    def __len__(self) -> int:
        return len(self.edges)

    def __getitem__(self, item: str) -> dict[str, Any] | None:
        return (self.data or {})[item]

    def __contains__(self, item: str) -> bool:
        return item not in self.json

    @property
    def queries(self) -> list[str]:
        """Return a list of the database queries that were executed."""
        return self.query_data.queries

    @property
    def query_log(self) -> str:
        """Return a string representation of the database queries that were executed."""
        return self.query_data.log

    @property
    def data(self) -> dict[str, Any] | None:
        """Return the data from the response content."""
        return self.json["data"]

    @property
    def first_query_object(self) -> dict[str, Any] | list[Any] | None:
        """
        Return the first query object in the response content.

        >>> self.json = {"data": {"foo": {"name": "bar"}}}
        >>> self.first_query_object
        {"name": "bar"}
        """
        data = self.data or {}
        try:
            return next(iter(data.values()))
        except StopIteration:  # pragma: no cover
            msg = f"No query object not found in response content: {self.json}"
            pytest.fail(msg, pytrace=False)

    @property
    def edges(self) -> list[dict[str, Any]]:
        """
        Return edges from the first query in the response content.

        >>> self.json = {"data": {"foo": {"edges": [{"node": {"name": "bar"}}]}}}
        >>> self.edges
        [{"node": {"name": "bar"}}]
        """
        try:
            return self.first_query_object["edges"]
        except (KeyError, TypeError):  # pragma: no cover
            msg = f"Edges not found in response content: {self.json}"
            pytest.fail(msg, pytrace=False)

    def node(self, index: int = 0) -> dict[str, Any]:
        """
        Return the node at the given index in the response content edges.

        >>> self.json = {"data": {"foo": {"edges": [{"node": {"name": "bar"}}]}}}
        >>> self.node(0)
        {"name": "bar"}
        """
        try:
            return self.edges[index]["node"]
        except (IndexError, TypeError):  # pragma: no cover
            msg = f"Node {index!r} not found in response content: {self.json}"
            pytest.fail(msg, pytrace=False)

    @property
    def has_errors(self) -> bool:  # pragma: no cover
        """Are there any errors in the response?"""
        return "errors" in self.json and self.json.get("errors") is not None

    @property
    def errors(self) -> list[dict[str, Any]]:
        """
        Return all errors.

        >>> self.json = {"errors": [{"message": "bar", "path": [...], ...}]}
        >>> self.errors
        [{"message": "bar", "path": [...], ...}]
        """
        try:
            return self.json["errors"]
        except (KeyError, TypeError):  # pragma: no cover
            msg = f"Errors not found in response content: {self.json}"
            pytest.fail(msg, pytrace=False)

    def error_message(self, selector: int | str = 0) -> str:  # pragma: no cover
        """
        Return the error message from the errors list...

        1) in the given index

        >>> self.json = {"errors": [{"message": "bar", ...}]}
        >>> self.error_message(0)  # default
        "bar"

        2) in the given path:

        >>> self.json = {"errors": [{"message": "bar", "path": ["fizz", "buzz", "foo"], ...}]}
        >>> self.error_message("foo")
        "bar"
        """
        if isinstance(selector, int):
            try:
                return self.errors[selector]["message"]
            except IndexError:
                msg = f"Errors list doesn't have an index {selector}: {self.json}"
                pytest.fail(msg, pytrace=False)
            except (KeyError, TypeError):
                msg = f"Field 'message' not found in error content: {self.json}"
                pytest.fail(msg, pytrace=False)
        else:
            try:
                return next(error["message"] for error in self.errors if error["path"][-1] == selector)
            except StopIteration:
                msg = f"Errors list doesn't have an error for field '{selector}': {self.json}"
                pytest.fail(msg, pytrace=False)
            except (KeyError, TypeError):
                msg = f"Field 'message' not found in error content: {self.json}"
                pytest.fail(msg, pytrace=False)

    @property
    def field_errors(self) -> list[FieldError]:
        """
        Return all field error messages.

        >>> self.json = {
        ...     "errors": [
        ...         {
        ...             "extensions": {
        ...                 "errors": [
        ...                     {
        ...                         "field": "foo",
        ...                         "messages": ["bar"],
        ...                         "codes": ["baz"],
        ...                     },
        ...                 ],
        ...             },
        ...         },
        ...     ],
        ... }
        ...
        >>> self.field_errors
        [{"field": "foo", "messages": ["bar"], "codes": ["baz"]}]
        """
        try:
            return [error for item in self.errors for error in item.get("extensions", {}).get("errors", [])]
        except (KeyError, TypeError):  # pragma: no cover
            msg = f"Field errors not found in response content: {self.json}"
            pytest.fail(msg, pytrace=False)

    def field_error_messages(self, field: str = "nonFieldErrors") -> list[str]:  # pragma: no cover
        """
        Return field error messages for desired field.

        >>> self.json = {
        ...     "errors": [
        ...         {
        ...             "extensions": {
        ...                 "errors": [
        ...                     {
        ...                         "field": "foo",
        ...                         "messages": "bar",
        ...                         "codes": "baz",
        ...                     },
        ...                     {
        ...                         "field": "foo",
        ...                         "messages": "one",
        ...                         "codes": "",
        ...                     },
        ...                 ],
        ...             },
        ...         },
        ...     ],
        ... }
        ...
        >>> self.field_error_messages("foo")
        ["bar", "one"]
        """
        messages: list[str] = []
        for error in self.field_errors:
            if error.get("field") == field:
                try:
                    messages.append(error["message"])
                except (KeyError, TypeError):
                    msg = f"Error message for field {field!r} not found in error: {error}"
                    pytest.fail(msg, pytrace=False)

        return messages

    def error_code(self, selector: int | str = 0) -> str:  # pragma: no cover
        """
        Return the error code from the errors list.

        1) in the given index

        >>> self.json = {"errors": [{"extensions": {"code": "bar"}, ...}]}
        >>> self.error_code(0)  # default
        "bar"

        2) in the given path:

        >>> self.json = {"errors": [{"extensions": {"code": "bar"}, "path": ["fizz", "buzz", "foo"], ...}]}
        >>> self.error_code("foo")
        "bar"
        """
        if isinstance(selector, int):
            try:
                return self.errors[selector]["extensions"]["code"]
            except (KeyError, TypeError):
                msg = f"Error code not found in error content: {self.json}"
                pytest.fail(msg, pytrace=False)
        else:
            try:
                return next(error["extensions"]["code"] for error in self.errors if error["path"][-1] == selector)
            except StopIteration:
                msg = f"Errors list doesn't have an error for field '{selector}': {self.json}"
                pytest.fail(msg, pytrace=False)
            except (KeyError, TypeError):
                msg = f"Field 'extensions' not found in error content: {self.json}"
                pytest.fail(msg, pytrace=False)

    def field_error_codes(self, field: str = "nonFieldErrors") -> list[str]:  # pragma: no cover
        """
        Return field error codes for desired field.

        >>> self.json = {
        ...     "errors": [
        ...         {
        ...             "extensions": {
        ...                 "errors": [
        ...                     {
        ...                         "field": "foo",
        ...                         "messages": "bar",
        ...                         "codes": "baz",
        ...                     },
        ...                     {
        ...                         "field": "foo",
        ...                         "messages": "one",
        ...                         "codes": "",
        ...                     },
        ...                 ],
        ...             },
        ...         },
        ...     ],
        ... }
        ...
        >>> self.field_error_codes("foo")
        ["bar", ""]
        """
        codes: list[str] = []
        for error in self.field_errors:
            if error.get("field") == field:
                try:
                    codes.append(error["code"])
                except (KeyError, TypeError):
                    msg = f"Error code for field {field!r} not found in error: {error}"
                    pytest.fail(msg, pytrace=False)

        return codes

    def assert_query_count(self, count: int) -> None:  # pragma: no cover
        if len(self.queries) != count:
            msg = f"Expected {count} queries, got {len(self.queries)}.\n{self.query_log}"
            pytest.fail(msg, pytrace=False)


class GraphQLClient(Client):
    response_class: ClassVar[type[GQLResponse]] = GQLResponse

    def __call__(  # noqa: PLR0913
        self: Self,
        query: str,
        input_data: dict[str, Any] | None = None,
        variables: dict[str, Any] | None = None,
        headers: dict[str, str] | None = None,
        operation_name: str | None = None,
    ) -> GQLResponse:
        """
        Make a GraphQL query to the test client.

        :params query: GraphQL query string.
        :params input_data: Set (and override) the "input" variable in the given variables.
        :params variables: Variables for the query.
        :params headers: Headers for the query.
        :params operation_name: Name of the operation to execute.
        """
        variables: dict[str, Any] = variables or {}
        if input_data is not None:
            variables.update({"input": input_data})

        files = extract_files(variables, prefix="variables")

        path_map: dict[str, list[str]] = {}
        files_map: dict[str, File] = {}
        for i, (file, path) in enumerate(files.items()):
            path_map[str(i)] = path
            files_map[str(i)] = file

        body: dict[str, Any] = {"query": query}
        if variables:
            body["variables"] = variables
        if operation_name is not None:  # pragma: no cover
            body["operationName"] = operation_name

        data = json.dumps(body)
        if files:
            data = {
                "operations": data,
                "map": json.dumps(path_map),
                **files_map,
            }

        with capture_database_queries() as results:
            response: HttpResponse = self.post(  # type: ignore[assignment]
                path=graphene_settings.TESTING_ENDPOINT,
                data=data,
                content_type=MULTIPART_CONTENT if files else "application/json",
                headers=headers,
            )

        return self.response_class(response, results)

    def login_with_superuser(self) -> User:
        user = get_user_model().objects.create_superuser(username="superuser", email="superuser@django.com")
        self.force_login(user)
        return user

    def login_with_regular_user(self) -> User:
        user = get_user_model().objects.create_user(username="user", email="user@django.com")
        self.force_login(user)
        return user



from .bases import CreateMutation, DeleteMutation, DjangoNode, UpdateMutation
from .filters import ModelFilterSet
from .serializers import NestingModelSerializer
from .views import FileUploadGraphQLView

__all__ = [
    "CreateMutation",
    "DeleteMutation",
    "DjangoNode",
    "FileUploadGraphQLView",
    "ModelFilterSet",
    "NestingModelSerializer",
    "UpdateMutation",
]



from __future__ import annotations

import json
import uuid
from dataclasses import dataclass, field
from enum import Enum
from itertools import chain

from django.db.models import lookups
from django.db.models.constants import LOOKUP_SEP
from graphene.utils.str_converters import to_camel_case

from ..settings import gdx_settings
from ..typing import Any, NamedTuple
from ..utils import get_nested

__all__ = [
    "build_query",
    "build_mutation",
]


class FiltersAndFields(NamedTuple):
    query_filters: str
    fields: str


@dataclass
class FieldFilterParams:
    filters: list[str] = field(default_factory=list)
    sub_filters: dict[str, FieldFilterParams] = field(default_factory=dict)

    def __str__(self) -> str:
        return f'({", ".join(self.filters)})'


KNOWN_LOOKUP_FIELDS: set[str] = {
    val.lookup_name
    for val in chain(
        lookups.__dict__.values(),
    )
    if getattr(val, "lookup_name", None) is not None
}


def build_query(__name: str, *, fields: str = "pk", connection: bool = False, **filter_params: Any) -> str:
    """
    Build a GraphQL query with the given field selections and filter parameters.

    :param __name: Name of the query, in camelCase.
    :param fields: Field selections as a GraphQL string.
    :param connection: Whether to build a Relay Connection query or basic one.
    :param filter_params: Parameters to use in the query. Will be converted to camelCase.
                          Use "__" to add filters to fields instead of the query.
    """
    result = _build_filters(fields, **filter_params)
    fields = f"edges {{ node {{ {result.fields} }} }}" if connection else result.fields
    return f"query {{ {__name}{result.query_filters} {{ {fields} }} }}"


def build_mutation(__name: str, __mutation_class_name: str, *, fields: str = "pk") -> str:
    """
    Build a GraphqQL mutation with the given field selections.

    :param __name: Name of the mutation, in camelCase.
    :param __mutation_class_name: Name of the mutation ObjectType, in PascalCase.
    :param fields: Field selections as a GraphQL string.
    """
    return f"mutation {__name}($input: {__mutation_class_name}Input!) {{ {__name}(input: $input) {{ {fields} }} }}"


def _build_filters(fields: str, /, **filter_params: Any) -> FiltersAndFields:
    """
    Build GrpahQL filters for the query and fields based on given params.

    >>> _build_filters("foo bar { fizz buzz }", one=1, foo__two=2, bar__three=3)
    FiltersAndFields(query_filters="(one: 1)", fields="foo(two: 2) bar(three: 3) { fizz buzz }")
    """
    params: dict[str, Any] = {}
    field_filter_params: dict[str, Any] = {}
    for key, value in filter_params.items():
        plain_key, _ = _split_lookups(key)
        if len(plain_key.split(LOOKUP_SEP)) > 1:
            field_filter_params[key] = value
        else:
            params[to_camel_case(key)] = _format_value_for_filter(
                value,
                is_order_by=key == gdx_settings.ORDERING_FILTER_NAME,
            )

    query_filters: str = ""
    if params:
        query_filters = f"({', '.join(f'{key}: {value}' for key, value in params.items())})"
    if field_filter_params:
        field_dict = _fields_to_dict(fields)
        field_filters = _build_field_filter_params(field_filter_params)
        fields = _add_filters_to_fields(field_dict, field_filters)

    return FiltersAndFields(query_filters=query_filters, fields=fields)


def _split_lookups(key: str) -> tuple[str, str]:
    """
    Split known lookup fields from the given filter key.

    >>> _split_lookups("foo__bar__istartswith")
    ("foo__bar", "istartswith")
    >>> _split_lookups("foo__bar__baz")
    ("foo__bar__baz", "")
    """
    plain_key = key
    for lookup_field in KNOWN_LOOKUP_FIELDS:
        plain_key = plain_key.removesuffix(f"{LOOKUP_SEP}{lookup_field}")
        if plain_key != key:
            return plain_key, lookup_field
    return plain_key, ""


def _format_value_for_filter(value: Any, *, is_order_by: bool = False) -> str:
    """Format values for the GraphQL filters. For enums, use the enum value. Otherwise, format as json."""
    if is_order_by:  # using custom enum `order_by` values
        if isinstance(value, list):
            return f"[{', '.join(item for item in value)}]"
        return value
    if isinstance(value, Enum):
        return value.value
    if isinstance(value, list) and all(isinstance(item, Enum) for item in value):
        return f"[{', '.join(str(item.value) for item in value)}]"
    if isinstance(value, uuid.UUID):
        value = str(value)
    return json.dumps(value)


def _build_field_filter_params(field_filter_params: dict[str, Any]) -> dict[str, FieldFilterParams]:
    """
    Convert field filters

    >>> _build_field_filter_params({"foo__bar": 1, "foo__baz": 2, "foo__fizz__buzz": 4, "other__thing": "value"})
    {
        "foo": FieldFilterParams(
            filters=['bar: 1', 'baz: 2'],
            sub_filters={
                "fizz": FieldFilterParams(
                    filters=['buzz: 4'],
                    sub_filters=[],
                )
            },
        ),
        "other": FieldFilterParams(
            filters=['thing: "value"'],
            sub_filters=[],
        ),
    }
    """
    params: dict[str, FieldFilterParams] = {}

    for key, value in field_filter_params.items():
        plain_key, lookup_field = _split_lookups(key)
        parts = plain_key.split(LOOKUP_SEP)
        last_index = len(parts) - 1
        current_params = params.setdefault(to_camel_case(parts[0]), FieldFilterParams())

        for i, part in enumerate(parts[1:], start=1):
            if i != last_index:
                sub_params = current_params.sub_filters.setdefault(to_camel_case(part), FieldFilterParams())
                current_params = sub_params
                continue

            if lookup_field:
                part = f"{part}{LOOKUP_SEP}{lookup_field}"  # noqa: PLW2901

            current_params.filters.append(f"{to_camel_case(part)}: {_format_value_for_filter(value)}")

    return params


def _add_filters_to_fields(fields: dict[str, dict | None], field_filter_params: dict[str, FieldFilterParams]) -> str:
    """
    Add field filters to fields and build into a GraphQL fields select string.

    >>> a = {"foo": {"bar": None}}
    >>> b = {"foo": FieldFilterParams(filters=["filter: 1"])
    >>> _add_filters_to_fields(fields=a, field_filter_params=b)
    "foo(filter: 1) { bar }"
    """
    for field_name, params in field_filter_params.items():
        if field_name not in fields:
            msg = (
                f"Field filters '{params}' defined for field '{field_name}' "
                f"but not in selected fields: '{_dict_to_fields(fields)}'"
            )
            raise RuntimeError(msg)

        if params.sub_filters:
            _add_filters_to_fields(fields=fields[field_name] or {}, field_filter_params=params.sub_filters)

        filters = f"({', '.join(params.filters)})" if params.filters else ""
        fields[f"{field_name}{filters}"] = fields.pop(field_name)

    return _dict_to_fields(fields)


def _fields_to_dict(fields: str) -> dict[str, dict | None]:
    """
    Convert a GraphQL fields select string into a dict describing the selections.

    >>> _fields_to_dict("foo bar { baz }")
    {"foo": None, "bar": {"baz": None}}
    """
    fields_dict = current_dict = {}
    parents: list[str] = []
    previous_field: str = ""

    for field_name in (val for val in fields.replace("\n", "").split(" ") if val != ""):
        if field_name == "{":
            parents.append(previous_field)
            current_dict[previous_field] = current_dict = {}
        elif field_name == "}":
            parents.pop()
            current_dict = get_nested(fields_dict, *parents, default={})
        elif any(val in field_name for val in ("{", "}")):
            msg = f"Should include a space before or after the '{{' and/or '}}' characters: '{field_name}'"
            raise RuntimeError(msg)
        else:
            current_dict[to_camel_case(field_name)] = None

        previous_field = field_name

    return fields_dict


def _dict_to_fields(fields: dict[str, dict | None]) -> str:
    """
    Reverse of `field_filters` above.

    >>> _dict_to_fields({"foo": None, "bar": {"baz": None}})
    "foo bar { baz }"
    """
    return " ".join(key if value is None else f"{key} {{ {_dict_to_fields(value)} }}" for key, value in fields.items())



from __future__ import annotations

import pytest

__all__ = [
    "graphql",
    "mock_png",
    "query_counter",
]


@pytest.fixture()
def graphql():  # noqa: ANN201
    from graphene_django_extensions.testing import GraphQLClient

    return GraphQLClient()


@pytest.fixture()
def query_counter():  # pragma: no cover  # noqa: ANN201
    from graphene_django_extensions.testing.utils import capture_database_queries

    return capture_database_queries


@pytest.fixture()
def mock_png():  # pragma: no cover  # noqa: ANN201
    from graphene_django_extensions.testing.utils import create_mock_png

    return create_mock_png()



from __future__ import annotations

from enum import Enum
from functools import partial
from typing import TYPE_CHECKING

import graphene
from django.forms import Form, ModelForm
from graphene import Field, InputField, InputObjectType, Mutation
from graphene.types.resolver import attr_resolver
from graphene.types.utils import yank_fields_from_attrs
from graphene_django import DjangoObjectType
from graphene_django.converter import convert_django_field
from graphene_django.forms.mutation import fields_for_form
from graphene_django.registry import get_global_registry
from graphene_django.rest_framework.mutation import fields_for_serializer
from graphene_django.types import ALL_FIELDS
from graphene_django.utils import is_valid_django_model
from graphql_relay import to_global_id
from query_optimizer import DjangoConnectionField, DjangoListField, optimize_single
from query_optimizer.settings import optimizer_settings
from rest_framework.exceptions import ValidationError as SerializerValidationError
from rest_framework.fields import SerializerMethodField, get_attribute
from rest_framework.serializers import ListSerializer, ModelSerializer, Serializer

from .connections import Connection
from .converters import convert_form_fields_to_not_required, convert_serializer_fields_to_not_required
from .errors import GQLPermissionDeniedError, validation_errors_to_graphql_errors
from .fields import RelatedField
from .model_operations import get_model_lookup_field, get_object_or_404
from .options import DjangoMutationOptions, DjangoNodeOptions
from .permissions import AllowAny, BasePermission, restricted_field
from .settings import gdx_settings
from .typing import Fields, GQLFields, Sequence
from .utils import add_translatable_fields, get_filter_info

if TYPE_CHECKING:
    from django.db import models
    from graphene.relay.node import NodeField

    from .typing import Any, AnyUser, FieldNameStr, GQLInfo, Literal, PermCheck, Self


__all__ = [
    "DjangoNode",
    "CreateMutation",
    "UpdateMutation",
    "DeleteMutation",
]


class DjangoNode(DjangoObjectType):
    """
    Custom base class for GraphQL-types that are backed by a Django model.
    Adds the following features to all types that inherit it:

    - Makes the `graphene.relay.Node` interface the default interface for the type.

    - Adds the `totalCount` field the default Connection field for the type.

    - Adds the `filter_queryset` method that can be overridden to add additional filtering for both
      `get_queryset` and `get_node` query sets.

    - Adds convenience methods for creating fields/list-fields/nodes/connections for the type.

    - Adds the `pk` field and resolver to the type if present on Meta.fields.

    ---

    The following options can be set in the `Meta`-class.

    `model: type[Model]`

    - Required. The model class for the node.

    `fields: list[FieldNameStr] | Literal["__all__"]`

    - Required. The fields to include in the node. If `__all__` is used, all fields are included.

    `permission_classes: Sequence[type[BasePermission]]`

    - Optional. Set permission classes for the node. Defaults to (`AllowAny`,).

    `restricted_fields: dict[FieldNameStr, PermCheck]`

    - Optional. Adds permission checks to the resolvers of the fields as defined in the dict.
    """

    _meta: DjangoNodeOptions

    class Meta:
        abstract = True

    @classmethod
    def Field(cls, **kwargs: Any) -> Field:  # noqa: N802
        return Field(cls, **kwargs)

    @classmethod
    def RelatedField(cls, **kwargs: Any) -> RelatedField:  # noqa: N802
        return RelatedField(cls, **kwargs)

    @classmethod
    def ListField(cls, **kwargs: Any) -> DjangoListField:  # noqa: N802
        return DjangoListField(cls, **kwargs)

    @classmethod
    def Node(cls, **kwargs: Any) -> NodeField:  # noqa: N802
        return graphene.relay.Node.Field(cls, **kwargs)

    @classmethod
    def Connection(cls, **kwargs: Any) -> DjangoConnectionField:  # noqa: N802
        return DjangoConnectionField(cls, **kwargs)  # pragma: no cover

    @classmethod
    def filter_queryset(cls, queryset: models.QuerySet, info: GQLInfo) -> models.QuerySet:
        """Implement this method filter to the available rows from the model on this node."""
        return queryset

    @classmethod
    def __init_subclass_with_meta__(  # noqa: PLR0913
        cls,
        model: type[models.Model] | None = None,
        fields: Fields | None = None,
        permission_classes: Sequence[type[BasePermission]] = (AllowAny,),
        restricted_fields: dict[FieldNameStr, PermCheck] | None = None,
        max_complexity: int | None = None,
        **options: Any,
    ) -> None:
        if model is None:  # pragma: no cover
            msg = "`Meta.model` is required."
            raise ValueError(msg)

        if not is_valid_django_model(model):  # pragma: no cover
            msg = f"`Meta.model` needs to be a Model-class, received: `{model}`."
            raise TypeError(msg)

        if fields is None:  # pragma: no cover
            msg = "`Meta.fields` is required."
            raise ValueError(msg)

        if not (fields == ALL_FIELDS or isinstance(fields, Sequence)):  # pragma: no cover
            msg = f"`Meta.fields` is should be a Sequence of field names or `{ALL_FIELDS}`, received: `{fields}`."
            raise TypeError(msg)

        if fields != ALL_FIELDS:
            fields = add_translatable_fields(model, fields)

        if not hasattr(cls, "pk") and (fields == ALL_FIELDS or "pk" in fields):
            cls._add_pk_field(model)

        if restricted_fields is not None:
            cls._add_field_restrictions(fields, restricted_fields)

        _meta = DjangoNodeOptions(
            class_type=cls,
            max_complexity=max_complexity or optimizer_settings.MAX_COMPLEXITY,
            permission_classes=permission_classes,
        )
        options.setdefault("connection_class", Connection)
        options.setdefault("interfaces", (graphene.relay.Node,))

        filterset_class = options.get("filterset_class", None)
        filter_fields: dict[str, list[str]] | None = options.pop("filter_fields", None)

        if filterset_class is None and filter_fields is not None:  # pragma: no cover
            from query_optimizer.filter import create_filterset

            options["filterset_class"] = create_filterset(model, filter_fields)

        super().__init_subclass_with_meta__(_meta=_meta, model=model, fields=fields, **options)

    @classmethod
    def _add_pk_field(cls, model: type[models.Model]) -> None:
        cls.pk = graphene.Int() if model._meta.pk.name == "id" else graphene.ID()
        cls.resolve_pk = cls.resolve_id

    @classmethod
    def _add_field_restrictions(cls, fields: Fields, restricted_fields: dict[FieldNameStr, PermCheck]) -> None:
        for field_name, check in restricted_fields.items():
            if fields != ALL_FIELDS and field_name not in fields:  # pragma: no cover
                msg = f"Field `{field_name}` not in `Meta.fields`."
                raise ValueError(msg)

            resolver = getattr(cls, f"resolve_{field_name}", None)
            if resolver is None:
                resolver = partial(attr_resolver, field_name, None)  # must be positional args!

            setattr(cls, f"resolve_{field_name}", restricted_field(check)(resolver))

    @classmethod
    def get_queryset(cls, queryset: models.QuerySet, info: GQLInfo) -> models.QuerySet:
        """Override `filter_queryset` instead of this method to add filtering of possible rows."""
        if not cls.has_filter_permissions(info):
            raise GQLPermissionDeniedError(
                message=gdx_settings.FILTER_PERMISSION_ERROR_MESSAGE,
                code=gdx_settings.FILTER_PERMISSION_ERROR_CODE,
            )
        return queryset

    @classmethod
    def get_node(cls, info: GQLInfo, pk: Any) -> models.Model | None:
        """Override `filter_queryset` instead of this method to add filtering of possible rows."""
        queryset = cls._meta.model._default_manager.all()
        instance = optimize_single(queryset, info, pk=pk, max_complexity=cls._meta.max_complexity)
        if instance is not None and not cls.has_node_permissions(info, instance):
            raise GQLPermissionDeniedError(
                message=gdx_settings.QUERY_PERMISSION_ERROR_MESSAGE,
                code=gdx_settings.QUERY_PERMISSION_ERROR_CODE,
            )
        return instance

    @classmethod
    def has_node_permissions(cls, info: GQLInfo, instance: models.Model) -> bool:
        """Check which permissions are required to access single items of this type."""
        filters = get_filter_info(info)
        return all(
            perm.has_node_permission(
                instance=instance,
                user=info.context.user,
                filters=filters,
            )
            for perm in cls._meta.permission_classes
        )

    @classmethod
    def has_filter_permissions(cls, info: GQLInfo) -> bool:
        """Check which permissions are required to access lists of this type."""
        filters = get_filter_info(info)
        return all(
            perm.has_filter_permission(
                user=info.context.user,
                filters=filters,
            )
            for perm in cls._meta.permission_classes
        )

    @classmethod
    def get_global_id(cls, pk: Any) -> str:
        """Get id used for `node` queries."""
        return to_global_id(cls.__name__, pk)


class DjangoMutation(Mutation):
    """
    Custom base class for GraphQL-mutations that are backed by a Django model.
    Adds the following features to all types that inherit it:

    - For updates, converts all fields to optional fields, enabling partial updates.

    - Checks for missing object types for nested model serializer fields to avoid nebulous import order errors.

    - Can add permission checks via permission classes.

    - Converts raise serializer validation errors to GraphQL validation errors.

    ---

    The following options can be set in the `Meta`-class.

    `model: type[models.Model]`

    - Required (delete only). Model class for the model the operation is performed on.

    `serializer_class: type[ModelSerializer | Serializer]`

    - Required (create and update only). The serializer used for the mutation.
      Type should be ModelSerializer for update and create operations, and Serializer for custom ones.

    `output_serializer_class: type[ModelSerializer | Serializer]`

    - Optional. The serializer used for the output data. If not set, `serializer_class` is used.
      Type should be ModelSerializer for update and create operations, and Serializer for custom ones.
      Serializer fields are modified so that all fields are optional, enabling partial updates.

    `permission_classes: Sequence[type[BasePermission]]`

    - Optional. Set permission classes for the mutation. Defaults to (`AllowAny`,).

    `lookup_field: str`

    - Optional. The field used for looking up the instance to be mutated. Defaults to the object's
      primary key, which is usually `id`. Note that the `lookup_field` attribute has to be available
      from the serializer's `Meta.fields` definition.

    `form_class: type[ModelForm | Form]`

    - Optional. Can be used instead of `serializer_class`.
      Type should be ModelForm for update and create operations, and Form for custom ones.

    `output_form_class: type[ModelForm | Form]`

    - Optional. Can be used instead of `output_serializer_class`.
      Type should be ModelForm for update and create operations, and Form for custom ones.
    """

    _meta: DjangoMutationOptions

    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(  # noqa: PLR0913
        cls,
        lookup_field: FieldNameStr | None = None,
        model: type[models.Model] | None = None,
        serializer_class: type[ModelSerializer | Serializer] | None = None,
        output_serializer_class: type[ModelSerializer | Serializer] | None = None,
        form_class: type[ModelForm | Form] | None = None,
        output_form_class: type[ModelForm | Form] | None = None,
        permission_classes: Sequence[type[BasePermission]] = (AllowAny,),
        model_operation: Literal["create", "update", "delete", "custom"] = "custom",
        **options: Any,
    ) -> None:
        _meta = DjangoMutationOptions(
            class_type=cls,
            lookup_field=lookup_field,
            model_class=model,
            serializer_class=serializer_class,
            output_serializer_class=output_serializer_class,
            form_class=form_class,
            output_form_class=output_form_class,
            permission_classes=permission_classes,
            model_operation=model_operation,
        )

        if model_operation in ("create", "update"):
            if serializer_class is not None:
                return cls.__init_subclass_update_or_create__(
                    _meta,
                    lookup_field,
                    serializer_class,
                    output_serializer_class,
                    model_operation,  # type: ignore[arg-type]
                    **options,
                )
            if form_class is not None:
                return cls.__init_subclass_form__(
                    _meta,
                    ModelForm,
                    form_class,
                    output_form_class,
                    **options,
                )

            msg = (  # pragma: no cover
                "`Meta.serializer_class` or `Meta.form_class` is required for create and update mutations."
            )
            raise ValueError(msg)  # pragma: no cover

        if model_operation == "delete":
            return cls.__init_subclass_delete__(
                _meta,
                lookup_field,
                model,
                **options,
            )

        if cls.custom_mutation == DjangoMutation.custom_mutation:  # pragma: no cover
            msg = "`custom_mutation` must be overridden in subclasses for custom mutations."
            raise ValueError(msg)

        if serializer_class is not None:
            return cls.__init_subclass_custom__(
                _meta,
                serializer_class,
                output_serializer_class,
                **options,
            )
        if form_class is not None:
            return cls.__init_subclass_form__(
                _meta,
                Form,
                form_class,
                output_form_class,
                **options,
            )

        msg = "`Meta.serializer_class` or `Meta.form_class` is required for custom mutations."  # pragma: no cover
        raise ValueError(msg)  # pragma: no cover

    @classmethod
    def __init_subclass_update_or_create__(
        cls,
        _meta: DjangoMutationOptions,
        lookup_field: FieldNameStr | None,
        serializer_class: type[ModelSerializer],
        output_serializer_class: type[ModelSerializer] | None,
        model_operation: Literal["create", "update"],
        **options: Any,
    ) -> None:
        if not issubclass(serializer_class, ModelSerializer):  # pragma: no cover
            msg = "`Meta.serializer_class` needs to be a ModelSerializer subclass."
            raise TypeError(msg)

        model = serializer_class.Meta.model
        lookup_field = get_model_lookup_field(model, lookup_field)

        if model_operation == "update":
            serializer_class = convert_serializer_fields_to_not_required(serializer_class, lookup_field)

        output_serializer_class = output_serializer_class or serializer_class
        if not issubclass(output_serializer_class, ModelSerializer):  # pragma: no cover
            msg = "`Meta.output_serializer_class` needs to be a ModelSerializer subclass."
            raise TypeError(msg)

        serializer = serializer_class()
        output_serializer = output_serializer_class()

        _check_serializer_field_models_in_registry(cls, output_serializer)

        input_fields: GQLFields = fields_for_serializer(
            serializer,
            only_fields=(),
            exclude_fields=(),
            is_input=True,
            lookup_field=lookup_field,
        )

        output_fields: GQLFields = fields_for_serializer(
            output_serializer,
            only_fields=(),
            exclude_fields=(),
            lookup_field=lookup_field,
        )

        _meta.lookup_field = lookup_field
        _meta.model_class = model
        _meta.serializer_class = serializer_class
        _meta.output_serializer_class = output_serializer_class

        return cls.__finish_init_subclass__(_meta, input_fields, output_fields, **options)

    @classmethod
    def __init_subclass_delete__(
        cls,
        _meta: DjangoMutationOptions,
        lookup_field: FieldNameStr | None,
        model: type[models.Model] | None,
        **options: Any,
    ) -> None:
        if model is None:  # pragma: no cover
            msg = "`Meta.model` is required."
            raise ValueError(msg)

        lookup_field = get_model_lookup_field(model, lookup_field)
        input_field = graphene.ID(required=True)
        if lookup_field != "pk":  # pragma: no cover
            input_field = convert_django_field(model._meta.get_field(lookup_field))

        input_fields: GQLFields = {lookup_field: input_field}
        output_fields: GQLFields = {"deleted": graphene.Boolean()}

        _meta.lookup_field = lookup_field

        return cls.__finish_init_subclass__(_meta, input_fields, output_fields, **options)

    @classmethod
    def __init_subclass_custom__(
        cls,
        _meta: DjangoMutationOptions,
        serializer_class: type[Serializer],
        output_serializer_class: type[Serializer] | None,
        **options: Any,
    ) -> None:
        if not issubclass(serializer_class, Serializer):  # pragma: no cover
            msg = "`Meta.serializer_class` needs to be a Serializer subclass."
            raise TypeError(msg)

        output_serializer_class = output_serializer_class or serializer_class
        if not issubclass(output_serializer_class, Serializer):  # pragma: no cover
            msg = "`Meta.output_serializer_class` needs to be a Serializer subclass."
            raise TypeError(msg)

        serializer = serializer_class()
        output_serializer = output_serializer_class()

        input_fields: GQLFields = fields_for_serializer(serializer, only_fields=(), exclude_fields=(), is_input=True)
        output_fields: GQLFields = fields_for_serializer(output_serializer, only_fields=(), exclude_fields=())

        _meta.serializer_class = serializer_class
        _meta.output_serializer_class = output_serializer_class

        return cls.__finish_init_subclass__(_meta, input_fields, output_fields, **options)

    @classmethod
    def __init_subclass_form__(
        cls,
        _meta: DjangoMutationOptions,
        required_form_class: type[ModelForm | Form],
        form_class: type[ModelForm, Form],
        output_form_class: type[ModelForm, Form] | None,
        **options: Any,
    ) -> None:
        if not issubclass(form_class, required_form_class):  # pragma: no cover
            msg = f"`Meta.form_class` needs to be a {required_form_class.__name__} subclass."
            raise TypeError(msg)

        output_form_class = output_form_class or form_class
        if not issubclass(output_form_class, required_form_class):  # pragma: no cover
            msg = f"`Meta.output_form_class` needs to be a {required_form_class.__name__} subclass."
            raise TypeError(msg)

        form = form_class()
        output_form_class = convert_form_fields_to_not_required(output_form_class)
        output_form = output_form_class()

        input_fields: GQLFields = fields_for_form(form, only_fields=(), exclude_fields=())
        output_fields: GQLFields = fields_for_form(output_form, only_fields=(), exclude_fields=())

        _meta.form_class = form_class
        _meta.output_form_class = output_form_class

        return cls.__finish_init_subclass__(_meta, input_fields, output_fields, **options)

    @classmethod
    def __finish_init_subclass__(
        cls,
        _meta: DjangoMutationOptions,
        input_fields: GQLFields,
        output_fields: GQLFields,
        **options: Any,
    ) -> None:
        options["name"] = cls.get_mutation_name(**options)

        arguments = cls.get_input_arguments(input_fields, **options)
        _meta.fields = yank_fields_from_attrs(attrs=output_fields, _as=Field)

        super().__init_subclass_with_meta__(_meta=_meta, arguments=arguments, **options)

    @classmethod
    def get_mutation_name(cls, **options: Any) -> str:
        base_name = options.get("name", cls.__name__).removesuffix("Payload")
        return f"{base_name}Payload"

    @classmethod
    def get_input_arguments(cls, input_fields: GQLFields, **options: Any) -> dict[str, InputObjectType]:
        input_fields = yank_fields_from_attrs(attrs=input_fields, _as=InputField)

        input_class: type[InputObjectType] = options.pop("input_class", InputObjectType)
        if not issubclass(input_class, InputObjectType):  # pragma: no cover
            msg = "`input_class` must be a subclass of `graphene.InputObjectType`."
            raise TypeError(msg)

        name = cls.get_mutation_name(**options).removesuffix("Payload")
        new_input_class = type(f"{name}Input", (input_class,), input_fields)
        return {"input": new_input_class(required=True)}

    @classmethod
    @validation_errors_to_graphql_errors
    def mutate(cls, root: Any, info: GQLInfo, **kwargs: Any) -> Self:
        input_data: dict[str, Any] = kwargs["input"]

        if cls._meta.model_operation == "delete":
            return cls.delete(info, input_data)
        if cls._meta.model_operation in ("create", "update"):
            return cls.update_or_create(info, input_data)

        validated_data = cls.validate_custom_mutation_input_data(info, input_data)
        return cls.custom_mutation(info, validated_data)

    @classmethod
    def update_or_create(cls, info: GQLInfo, input_data: dict[str, Any]) -> Self:
        maybe_instance = cls.get_instance(input_data)

        if cls._meta.model_operation == "create" and not cls.has_create_permissions(info, input_data):
            raise GQLPermissionDeniedError(
                message=gdx_settings.CREATE_PERMISSION_ERROR_MESSAGE,
                code=gdx_settings.CREATE_PERMISSION_ERROR_CODE,
            )
        if cls._meta.model_operation == "update" and not cls.has_update_permissions(maybe_instance, info, input_data):
            raise GQLPermissionDeniedError(
                message=gdx_settings.UPDATE_PERMISSION_ERROR_MESSAGE,
                code=gdx_settings.UPDATE_PERMISSION_ERROR_CODE,
            )

        validator = cls.run_validation(input_data, info, maybe_instance)
        instance = validator.save()

        if cls._meta.serializer_class is not None:
            output = cls.get_serializer_output(instance)
        else:
            output = cls.get_form_output(instance)

        return cls(**output)  # type: ignore[arg-type]

    @classmethod
    def delete(cls, info: GQLInfo, input_data: dict[str, Any]) -> Self:
        instance = cls.get_instance(input_data)
        if not cls.has_delete_permissions(instance, info, input_data):
            raise GQLPermissionDeniedError(
                message=gdx_settings.DELETE_PERMISSION_ERROR_MESSAGE,
                code=gdx_settings.DELETE_PERMISSION_ERROR_CODE,
            )

        cls.validate_deletion(instance, info.context.user)
        count, _ = instance.delete()
        return cls(deleted=bool(count))  # type: ignore[arg-type]

    @classmethod
    def custom_mutation(cls, info: GQLInfo, input_data: dict[str, Any]) -> Self:  # pragma: no cover
        """Override this method to perform custom mutations instead of the regular create, update or delete."""
        raise NotImplementedError(f"Custom model operation not defined for '{cls.__name__}'")  # noqa: EM102

    @classmethod
    def validate_custom_mutation_input_data(cls, info: GQLInfo, input_data: dict[str, Any]) -> dict[str, Any]:
        if not cls.has_mutation_permission(info, input_data):
            raise GQLPermissionDeniedError(
                message=gdx_settings.MUTATION_PERMISSION_ERROR_MESSAGE,
                code=gdx_settings.MUTATION_PERMISSION_ERROR_CODE,
            )

        validator = cls.run_validation(input_data, info)
        if isinstance(validator, Serializer):
            return validator.validated_data
        return validator.cleaned_data

    @classmethod
    def run_validation(
        cls,
        input_data: dict[str, Any],
        info: GQLInfo,
        maybe_instance: models.Model | None = None,
    ) -> Serializer | Form:
        validator_class: type[Serializer | Form] = cls._meta.serializer_class or cls._meta.form_class

        for key, maybe_enum in input_data.items():
            if isinstance(maybe_enum, Enum):
                input_data[key] = maybe_enum.value

        validator_kwargs: dict[str, Any] = {"data": input_data}
        if issubclass(validator_class, ModelForm):
            validator_kwargs["instance"] = maybe_instance
        if issubclass(validator_class, Serializer):
            validator_kwargs["instance"] = maybe_instance
            validator_kwargs["context"] = {"request": info.context}
            validator_kwargs["partial"] = maybe_instance is not None

        validator = validator_class(**validator_kwargs)
        if not validator.is_valid():
            raise SerializerValidationError(validator.errors)
        return validator

    @classmethod
    def get_instance(cls, input_data: dict[str, Any]) -> models.Model | None:
        if cls._meta.model_operation == "create":
            return None
        lookup_field: str = cls._meta.lookup_field
        if lookup_field not in input_data:  # pragma: no cover
            raise SerializerValidationError({lookup_field: "This field is required."})
        return get_object_or_404(cls._meta.model_class, **{lookup_field: input_data[lookup_field]})

    @classmethod
    def get_serializer_output(cls, instance: models.Model) -> dict[str, Any]:
        serializer = cls._meta.output_serializer_class()

        kwargs: dict[str, Any] = {}
        for field_name, field in serializer.fields.items():
            if not field.write_only:
                if isinstance(field, SerializerMethodField):  # pragma: no cover
                    kwargs[field_name] = field.to_representation(instance)
                else:
                    kwargs[field_name] = field.get_attribute(instance)
        return kwargs

    @classmethod
    def get_form_output(cls, instance: models.Model) -> dict[str, Any]:
        form = cls._meta.output_form_class()
        return {field_name: get_attribute(instance, [field_name]) for field_name in form.fields}

    @classmethod
    def validate_deletion(cls, instance: models.Model, user: AnyUser) -> None:
        """Implement to perform additional validation before the given instance is deleted."""
        return  # pragma: no cover

    @classmethod
    def has_mutation_permission(cls, info: GQLInfo, input_data: dict[str, Any]) -> bool:
        return all(
            perm.has_mutation_permission(
                user=info.context.user,
                input_data=input_data,
            )
            for perm in cls._meta.permission_classes
        )

    @classmethod
    def has_create_permissions(cls, info: GQLInfo, input_data: dict[str, Any]) -> bool:
        return all(
            perm.has_create_permission(
                user=info.context.user,
                input_data=input_data,
            )
            for perm in cls._meta.permission_classes
        )

    @classmethod
    def has_update_permissions(cls, instance: models.Model, info: GQLInfo, input_data: dict[str, Any]) -> bool:
        return all(
            perm.has_update_permission(
                instance=instance,
                user=info.context.user,
                input_data=input_data,
            )
            for perm in cls._meta.permission_classes
        )

    @classmethod
    def has_delete_permissions(cls, instance: models.Model, info: GQLInfo, input_data: dict[str, Any]) -> bool:
        return all(
            perm.has_delete_permission(
                instance=instance,
                user=info.context.user,
                input_data=input_data,
            )
            for perm in cls._meta.permission_classes
        )


class CreateMutation(DjangoMutation):
    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(cls, **options: Any) -> None:
        super().__init_subclass_with_meta__(model_operation="create", **options)


class UpdateMutation(DjangoMutation):
    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(cls, **options: Any) -> None:
        super().__init_subclass_with_meta__(model_operation="update", **options)


class DeleteMutation(DjangoMutation):
    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(cls, **options: Any) -> None:
        super().__init_subclass_with_meta__(model_operation="delete", **options)


def _check_serializer_field_models_in_registry(
    mutation_class: type[DjangoMutation],
    serializer: ModelSerializer,
) -> None:
    global_registry = get_global_registry()

    for field in serializer.fields.values():
        if isinstance(field, ModelSerializer):
            field_model: type[models.Model] = field.Meta.model
        elif isinstance(field, ListSerializer) and isinstance(field.child, ModelSerializer):
            field = field.child  # noqa: PLW2901
            field_model: type[models.Model] = field.Meta.model
        else:
            continue

        type_ = global_registry.get_type_for_model(field_model)
        if type_ is None:  # pragma: no cover
            # This error is in place since the mutation class needs to know the ObjectTypes
            # of the models of nested model serializer fields when converting them to ObjectTypes,
            # but currently does not properly warn against this error itself. See this line in
            # `graphene_django.rest_framework.serializer_converter.convert_serializer_field`:
            #
            # args = [global_registry.get_type_for_model(field_model)]
            #
            # Since model is not registered, `global_registry.get_type_for_model` returns `None`,
            # which will cause errors later on in the mutation class creation process.
            msg = (
                f"Could not find a ObjectType for model: `{field_model.__name__}`. "
                f"Make sure that the ObjectType for this model is registered before the "
                f"`{mutation_class.__name__}` mutation class is crated. This can be achieved by "
                f"importing the ObjectType before the mutation class is created."
            )
            raise LookupError(msg)

        _check_serializer_field_models_in_registry(mutation_class, field)



from __future__ import annotations

from functools import cache
from typing import TYPE_CHECKING

from graphene import Connection
from graphene.utils.str_converters import to_snake_case
from graphql import get_argument_values
from graphql.execution.execute import get_field_def
from graphql.language.ast import ExecutableDefinitionNode, FieldNode, SelectionNode
from query_optimizer.ast import get_underlying_type

from .constants import Operation
from .settings import gdx_settings
from .typing import GraphQLFilterInfo, StrEnum

if TYPE_CHECKING:
    from django.db import models
    from graphql import GraphQLField, GraphQLObjectType, GraphQLSchema

    from .typing import Any, GQLInfo, Iterable, Sequence

__all__ = [
    "get_fields_from_info",
    "get_filter_info",
    "get_nested",
    "add_translatable_fields",
    "get_operator_enum",
]


def get_nested(obj: dict | list | None, /, *args: str | int, default: Any = None) -> Any:
    """
    Get value from a nested structure containing dicts with string keys or lists,
    where the keys and list indices might not exist.

    1) `data["foo"][0]["bar"]["baz"]`
     - Might raise a `KeyError` or `IndexError` if any of the keys or indices don't exist.

    2) `get_nested(data, "foo", 0, "bar", "baz")`
     - Will return `None` (default) if any of the keys or indices don't exist.
    """
    if not args:
        return obj if obj is not None else default

    arg, args = args[0], args[1:]

    if isinstance(arg, int):
        obj = obj or []
        try:
            obj = obj[arg]
        except (IndexError, KeyError):
            obj = None
        return get_nested(obj, *args, default=default)

    obj = obj or {}
    try:
        return get_nested(obj.get(arg), *args, default=default)
    except AttributeError:
        obj = None
        return get_nested(obj, *args, default=default)


def get_filter_info(info: GQLInfo) -> dict[str, Any]:
    """Find filter arguments from the GraphQL query."""
    args = _get_arguments(info.field_nodes, info.variable_values, info.parent_type, info.schema)
    if not args:
        return {}
    return args[to_snake_case(info.field_name)]


def _get_arguments(
    field_nodes: Iterable[FieldNode | SelectionNode],
    variable_values: dict[str, Any],
    parent: GraphQLObjectType,
    schema: GraphQLSchema,
) -> dict[str, GraphQLFilterInfo]:
    arguments: dict[str, GraphQLFilterInfo] = {}
    for field_node in field_nodes:
        if not isinstance(field_node, FieldNode):  # pragma: no cover
            continue

        field_def: GraphQLField | None = get_field_def(schema, parent, field_node)
        if field_def is None:  # pragma: no cover
            continue

        name = to_snake_case(field_node.name.value)
        filters = get_argument_values(type_def=field_def, node=field_node, variable_values=variable_values)

        new_parent = get_underlying_type(field_def.type)

        # If the field is a connection, we need to go deeper to get the actual field
        if issubclass(getattr(new_parent, "graphene_type", type(None)), Connection):
            field_def = new_parent.fields["edges"]
            new_parent = get_underlying_type(field_def.type)
            field_def = new_parent.fields["node"]
            new_parent = get_underlying_type(field_def.type)
            field_node = field_node.selection_set.selections[0].selection_set.selections[0]  # noqa: PLW2901

        arguments[name] = info = GraphQLFilterInfo(name=new_parent.name, filters=filters, children={})

        if field_node.selection_set is not None:
            result = _get_arguments(field_node.selection_set.selections, variable_values, new_parent, schema)
            if result:
                info["children"] = result

    return {name: field for name, field in arguments.items() if field["filters"] or field["children"]}


def get_fields_from_info(info: GQLInfo) -> list[dict[str, Any]]:
    """Find selected fields from the GraphQL query and return them as a list of dicts."""
    return _get_field_node(info.operation)


def _get_field_node(field: ExecutableDefinitionNode | FieldNode) -> dict[str, Any] | list[Any]:
    filters: list[Any] = []
    for selection in field.selection_set.selections:
        if not isinstance(selection, FieldNode):  # pragma: no cover
            continue
        if selection.selection_set is not None:
            filters.append({selection.name.value: _get_field_node(selection)})
        else:
            filters.append(selection.name.value)
    return filters


def add_translatable_fields(model: type[models.Model], fields: Sequence[str]) -> Sequence[str]:  # pragma: no cover
    """
    If `django-modeltranslation` is installed, find and add all translation fields
    to the given fields list, for the given fields, in the given model.
    """
    try:
        from modeltranslation.manager import get_translatable_fields_for_model
        from modeltranslation.utils import get_translation_fields
    except ImportError:
        return fields

    translatable_fields: list[str] = get_translatable_fields_for_model(model) or []
    new_fields: list[str] = []
    for field in fields:
        new_fields.append(field)
        if field not in translatable_fields:
            continue
        fields = get_translation_fields(field)
        new_fields.extend(fields)

    return new_fields


@cache
def get_operator_enum() -> type[Operation]:
    if not gdx_settings.EXTEND_USER_DEFINED_FILTER_OPERATIONS:  # pragma: no cover
        return Operation

    current_members: dict[str, str] = {key: value.value for key, value in Operation._member_map_.items()}
    for operator in gdx_settings.EXTEND_USER_DEFINED_FILTER_OPERATIONS:
        current_members[operator.upper()] = operator.upper()

    return StrEnum(Operation.__name__, current_members)  # type: ignore[return-value]



from .form import (
    EnumChoiceField,
    EnumMultipleChoiceField,
    IntChoiceField,
    IntMultipleChoiceField,
    OrderByField,
    UserDefinedFilterField,
)
from .graphql import (
    DjangoConnectionField,
    DjangoListField,
    Duration,
    OrderingChoices,
    RelatedField,
    Time,
    TypedDictField,
    TypedDictListField,
    UserDefinedFilterInputType,
)
from .serializer import DurationField, EnumFriendlyChoiceField, IntegerPrimaryKeyField

__all__ = [
    "DjangoConnectionField",
    "DjangoListField",
    "Duration",
    "DurationField",
    "EnumChoiceField",
    "EnumFriendlyChoiceField",
    "EnumMultipleChoiceField",
    "IntChoiceField",
    "IntegerPrimaryKeyField",
    "IntMultipleChoiceField",
    "OrderByField",
    "OrderingChoices",
    "RelatedField",
    "Time",
    "TypedDictField",
    "TypedDictListField",
    "UserDefinedFilterField",
    "UserDefinedFilterInputType",
]



from __future__ import annotations

import datetime
from enum import Enum
from typing import TYPE_CHECKING

from django.core.validators import MinValueValidator
from rest_framework import serializers
from rest_framework.relations import PKOnlyObject

if TYPE_CHECKING:
    from django.db.models import Model

    from ..typing import Any


__all__ = [
    "IntegerPrimaryKeyField",
    "IntPkOnlyObject",
    "DurationField",
    "EnumFriendlyChoiceField",
]


class IntPkOnlyObject(PKOnlyObject):
    """PK object that is coerced to an integer."""

    def __int__(self) -> int:
        return int(self.pk)


class IntegerPrimaryKeyField(serializers.PrimaryKeyRelatedField, serializers.IntegerField):
    """A field that refers to foreign keys by an integer primary key."""

    def get_attribute(self, instance: Model) -> IntPkOnlyObject | None:
        attribute = super().get_attribute(instance)
        if isinstance(attribute, PKOnlyObject) and attribute.pk:
            return IntPkOnlyObject(pk=attribute.pk)
        return None  # pragma: no cover


class EnumFriendlyChoiceField(serializers.ChoiceField):
    """ChoiceField that works with enum inputs as well."""

    def __init__(self, choices: list[tuple[Any, Any]], **kwargs: Any) -> None:
        # Enum can be provided as a keyword argument, used for mutation input enum naming.
        # See: `graphene_django_extensions.converters.convert_serializer_field_to_enum`
        self.enum: Enum | None = kwargs.pop("enum", None)
        super().__init__(choices, **kwargs)

    def to_internal_value(self, data: Any) -> str:
        if data == "" and self.allow_blank:  # pragma: no cover
            return ""
        if isinstance(data, Enum):  # pragma: no cover
            data = data.value
        try:
            return self.choice_strings_to_values[str(data)]
        except KeyError:  # pragma: no cover
            self.fail("invalid_choice", input=data)

    def to_representation(self, value: Any) -> Enum:  # pragma: no cover
        if value in ("", None):
            return value
        if isinstance(value, Enum):
            value = value.value
        return self.choice_strings_to_values.get(str(value), value)


class MinDurationValidator(MinValueValidator):
    def clean(self, x: datetime.timedelta) -> int:
        return int(x.total_seconds())


class DurationField(serializers.IntegerField):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.validators.append(MinDurationValidator(0))

    def to_internal_value(self, data: Any) -> datetime.timedelta:
        if isinstance(data, datetime.timedelta):
            return data
        if isinstance(data, int):
            return datetime.timedelta(seconds=data)
        try:  # pragma: no cover
            data = int(data)
        except (ValueError, TypeError):  # pragma: no cover
            self.fail("invalid")
        return datetime.timedelta(seconds=data)  # pragma: no cover

    def to_representation(self, value: datetime.timedelta) -> int:  # pragma: no cover
        return int(value.total_seconds())

    def get_attribute(self, instance: Any) -> int | None:
        value = super().get_attribute(instance)
        if isinstance(value, datetime.timedelta):
            return int(value.total_seconds())
        return value  # pragma: no cover



from .builders import build_mutation, build_query
from .client import GraphQLClient
from .utils import capture_database_queries, compare_unordered, create_mock_png, parametrize_helper

__all__ = [
    "GraphQLClient",
    "build_mutation",
    "build_query",
    "capture_database_queries",
    "compare_unordered",
    "create_mock_png",
    "parametrize_helper",
]



from __future__ import annotations

import dataclasses
from functools import wraps
from typing import TYPE_CHECKING

from django.db import IntegrityError, models, transaction
from django.db.models import NOT_PROVIDED
from graphene_django.types import ALL_FIELDS
from rest_framework.exceptions import ValidationError
from rest_framework.relations import ManyRelatedField, RelatedField
from rest_framework.serializers import ListSerializer, ModelSerializer

from .errors import get_constraint_message
from .fields import DurationField, EnumFriendlyChoiceField, IntegerPrimaryKeyField
from .model_operations import RelatedFieldInfo, get_object_or_404, get_related_field_info
from .typing import ParamSpec, SerializerMeta, TypeVar
from .utils import add_translatable_fields

if TYPE_CHECKING:
    from django.db.models import Model
    from rest_framework.fields import Field

    from .typing import Any, AnyUser, Callable, Self


__all__ = [
    "NestingModelSerializer",
]


@dataclasses.dataclass
class PreSaveInfo:
    field: Field
    initial_data: Any
    related_info: RelatedFieldInfo


class NotProvided:  # pragma: no cover
    def __bool__(self) -> bool:
        return False


NotProvided = NotProvided()


T = TypeVar("T")
P = ParamSpec("P")


def _related_pre_and_post_save(func: Callable[P, T]) -> Callable[P, T]:
    """Handle related models before and after creating or updating the main model."""

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> models.Model:
        self: NestingModelSerializer = args[0]
        validated_data = next((arg for arg in args if isinstance(arg, dict)), kwargs.get("validated_data"))
        if validated_data is None:  # pragma: no cover
            msg = "'validated_data' not found in args or kwargs"
            raise ValueError(msg)

        try:
            with transaction.atomic():
                related_serializers = self._pre_save(validated_data)
                instance = func(*args, **kwargs)
                if related_serializers:
                    self._post_save(instance, related_serializers)
        except IntegrityError as error:
            msg = get_constraint_message(error.args[0])
            raise ValidationError(msg) from error

        return instance

    return wrapper


class NestingModelSerializer(ModelSerializer):
    """
    ModelSerializer that contains logic for updating and creating related models
    when they are included as nested (list)serializer fields or (lists of) primary keys.
    """

    instance: Model  # Use this to hint the instance model type in subclasses
    serializer_related_field = IntegerPrimaryKeyField  # Related fields defined in Meta are integers by default
    serializer_choice_field = EnumFriendlyChoiceField  # Converts enums to string correctly
    serializer_field_mapping = ModelSerializer.serializer_field_mapping | {
        models.DurationField: DurationField,
    }

    class Meta(SerializerMeta):
        pass

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        if cls.Meta.fields != ALL_FIELDS:
            cls.Meta.fields = add_translatable_fields(cls.Meta.model, cls.Meta.fields)
        return super().__new__(cls, *args, **kwargs)

    def get_or_default(self, field: str, attrs: dict[str, Any]) -> Any:
        """
        Get field value from attrs, or if not found, use a default value from
        1) the serializer model instance, or 2) the model field.
        """
        default = self.Meta.model._meta.get_field(field).default
        default = getattr(self.instance, field, default)
        value = attrs.get(field, default)
        if value is NOT_PROVIDED:  # pragma: no cover
            return NotProvided  # This one is falsy
        return value

    @property
    def request_user(self) -> AnyUser:
        return self.context["request"].user

    def get_update_or_create(self, data: dict[str, Any] | None) -> Model | None:
        if data is None:  # pragma: no cover
            return None

        pk = data.pop("pk", None)
        if pk is not None:
            instance = get_object_or_404(self.Meta.model, pk=pk)
            if not data:
                return instance
            return self.update(instance, data)

        return self.create(data)

    @_related_pre_and_post_save
    def create(self, validated_data: dict[str, Any]) -> Model:
        """Create a new instance of the model, while also handling related models."""
        return self.Meta.model._default_manager.create(**validated_data)

    @_related_pre_and_post_save
    def update(self, instance: Model, validated_data: dict[str, Any]) -> Model:
        """Update an existing instance of the model, while also handling related models."""
        for attr, value in validated_data.items():
            setattr(instance, attr, value)
        instance.save()
        return instance

    def _pre_save(self, validated_data: dict[str, Any]) -> list[PreSaveInfo]:
        """
        Prepare related models defined using BaseModelSerializers.
        Forward 'one-to-one' and 'many-to-one' related entities will be fetched, updated, or created.
        Other related entities will be saved to be handled after the main model is saved using '_handle_to_many'.
        """
        pre_save_infos: list[PreSaveInfo] = []
        related_info = get_related_field_info(self.Meta.model)

        for name in list(validated_data):  # Copy keys so that we can pop from the original dict in the loop
            field: Field | None = self.fields.get(name, None)
            if field is None:
                continue

            related_field_info = related_info.get(name, None)
            if related_field_info is None:
                continue

            if related_field_info.one_to_one or related_field_info.many_to_one:
                info = self._pre_handle_to_one(field, related_field_info, validated_data, name)
                if info is not None:
                    pre_save_infos.append(info)

            elif related_field_info.one_to_many or related_field_info.many_to_many:
                info = self._pre_handle_to_many(field, related_field_info, validated_data, name)
                if info is not None:
                    pre_save_infos.append(info)

        return pre_save_infos

    def _pre_handle_to_one(
        self,
        field: Field,
        related_info: RelatedFieldInfo,
        validated_data: dict[str, Any],
        key: str,
    ) -> PreSaveInfo | None:
        initial_data: Any = validated_data.pop(key, None)
        if related_info.reverse:
            return PreSaveInfo(field=field, initial_data=initial_data, related_info=related_info)
        if isinstance(field, NestingModelSerializer):
            validated_data[key] = field.get_update_or_create(initial_data)
        else:
            validated_data[key] = initial_data
        return None

    def _pre_handle_to_many(
        self,
        field: Field,
        related_info: RelatedFieldInfo,
        validated_data: dict[str, Any],
        key: str,
    ) -> PreSaveInfo | None:
        initial_data: Any = validated_data.pop(key, None)
        if initial_data is None:  # pragma: no cover
            return None
        return PreSaveInfo(field=field, initial_data=initial_data, related_info=related_info)

    def _post_save(self, instance: Model, pre_save_info: list[PreSaveInfo]) -> None:
        """
        Handle creating or updating related models after the main model.
        Delete any existing 'one_to_many' entities that were untouched in this request.
        Add any new 'many_to_many' entities that were not previously linked to the main model.
        """
        for info in pre_save_info:
            # Handle reverse one-to-one
            if info.related_info.one_to_one and info.related_info.reverse:
                self._post_handle_reverse_one_to_one(instance, info)
            elif info.related_info.one_to_many:
                self._post_handle_one_to_many(instance, info)
            elif info.related_info.many_to_many:
                self._post_handle_many_to_many(instance, info)

    def _post_handle_reverse_one_to_one(self, instance: Model, info: PreSaveInfo) -> None:
        if info.initial_data is None:  # pragma: no cover
            related_instance = getattr(instance, info.related_info.field_name, None)
            if related_instance is not None:
                setattr(related_instance, info.related_info.related_name, None)
                related_instance.save()
            return

        if isinstance(info.field, NestingModelSerializer):
            info.initial_data[info.related_info.related_name] = instance
            info.field.get_update_or_create(info.initial_data)
            return

        if isinstance(info.field, RelatedField):
            setattr(info.initial_data, info.related_info.related_name, instance)
            info.initial_data.save()
            return

    def _post_handle_one_to_many(self, instance: Model, info: PreSaveInfo) -> None:
        if isinstance(info.field, ListSerializer) and isinstance(info.field.child, NestingModelSerializer):
            pks: list[Any] = []

            for initial_data in info.initial_data:
                initial_data[info.related_info.related_name] = instance
                nested_instance = info.field.child.get_update_or_create(initial_data)
                if nested_instance is not None:
                    pks.append(nested_instance.pk)

            # Delete related objects that were not created or modified.
            selector = {info.related_info.related_name: instance}
            info.field.child.Meta.model.objects.filter(**selector).exclude(pk__in=pks).delete()
            return

        if isinstance(info.field, ManyRelatedField):
            getattr(instance, info.related_info.field_name).set(info.initial_data)
            return

    def _post_handle_many_to_many(self, instance: Model, info: PreSaveInfo) -> None:
        if isinstance(info.field, ListSerializer) and isinstance(info.field.child, NestingModelSerializer):
            instances: list[Model] = []

            for item in info.initial_data:
                nested_instance = info.field.child.get_update_or_create(item)
                if nested_instance is not None:
                    instances.append(nested_instance)

            # Add related objects that were not previously linked to the main model.
            getattr(instance, info.related_info.field_name).set(instances)
            return

        if isinstance(info.field, ManyRelatedField):
            getattr(instance, info.related_info.field_name).set(info.initial_data)
            return

    def get_fields(self) -> dict[str, Field]:
        fields = super().get_fields()
        # Add the model field enum to `EnumFriendlyChoiceField`
        # if the enum was not explicitly defined in the serializer field.
        # (e.g. field created automatically based on `serializer_choice_field`)
        # This is used for mutation input enum naming.
        for name, field in fields.items():
            if hasattr(field, "enum") and field.enum is None:
                model_field = self.Meta.model._meta.get_field(name)
                field.enum = getattr(model_field, "enum", None)
        return fields



from django.conf import settings
from django.db import models
from django.test.signals import setting_changed
from settings_holder import SettingsHolder, reload_settings

from .typing import Any, NamedTuple, Union

SETTING_NAME: str = "GRAPHENE_DJANGO_EXTENSIONS"


class DefaultSettings(NamedTuple):
    QUERY_PERMISSION_ERROR_MESSAGE: str = "No permission to access node."
    QUERY_PERMISSION_ERROR_CODE: str = "NODE_PERMISSION_DENIED"
    FILTER_PERMISSION_ERROR_MESSAGE: str = "No permission to access node."
    FILTER_PERMISSION_ERROR_CODE: str = "FILTER_PERMISSION_DENIED"
    FIELD_PERMISSION_ERROR_MESSAGE: str = "No permission to access field."
    FIELD_PERMISSION_ERROR_CODE: str = "FIELD_PERMISSION_DENIED"
    MUTATION_PERMISSION_ERROR_MESSAGE: str = "No permission to mutate."
    MUTATION_PERMISSION_ERROR_CODE: str = "MUTATION_PERMISSION_DENIED"
    CREATE_PERMISSION_ERROR_MESSAGE: str = "No permission to create."
    CREATE_PERMISSION_ERROR_CODE: str = "CREATE_PERMISSION_DENIED"
    UPDATE_PERMISSION_ERROR_MESSAGE: str = "No permission to update."
    UPDATE_PERMISSION_ERROR_CODE: str = "UPDATE_PERMISSION_DENIED"
    DELETE_PERMISSION_ERROR_MESSAGE: str = "No permission to delete."
    DELETE_PERMISSION_ERROR_CODE: str = "DELETE_PERMISSION_DENIED"
    MUTATION_VALIDATION_ERROR_MESSAGE: str = "Mutation was unsuccessful."
    MUTATION_VALIDATION_ERROR_CODE: str = "MUTATION_VALIDATION_ERROR"
    NOT_FOUND_ERROR_CODE: str = "NOT_FOUND"
    ORDERING_FILTER_NAME: str = "order_by"
    EXTEND_USER_DEFINED_FILTER_OPERATIONS: list[str] | None = None


DEFAULTS: dict[str, Any] = DefaultSettings()._asdict()

IMPORT_STRINGS: set[Union[bytes, str]] = set()
REMOVED_SETTINGS: set[str] = {
    "PERMISSION_DENIED_MESSAGE",
}

gdx_settings = SettingsHolder(
    setting_name=SETTING_NAME,
    defaults=DEFAULTS,
    import_strings=IMPORT_STRINGS,
    removed_settings=REMOVED_SETTINGS,
)

reload_my_settings = reload_settings(SETTING_NAME, gdx_settings)
setting_changed.connect(reload_my_settings)


# Override the default Enum name generator to have PascalCase names everywhere


def enum_name(field: models.Field) -> str:
    # If using `StrChoiceField` or `IntChoiceField`, use the stored enum name.
    if hasattr(field, "enum") and field.enum is not None:  # pragma: no cover
        return field.enum.__name__
    # Otherwise, generate the name from the field name.
    return "".join(s.capitalize() for s in field.name.split("_"))


if not hasattr(settings, "GRAPHENE"):  # pragma: no cover
    settings.GRAPHENE = {}

settings.GRAPHENE.setdefault("DJANGO_CHOICE_FIELD_ENUM_CUSTOM_NAME", "graphene_django_extensions.settings.enum_name")



from __future__ import annotations

import re
from functools import wraps
from typing import TYPE_CHECKING

from django.apps import apps
from django.core.exceptions import ValidationError as DjangoValidationError
from django.db import models
from graphene_django.settings import graphene_settings
from graphene_django.utils import camelize
from graphql import GraphQLError
from rest_framework.exceptions import ValidationError as SerializerValidationError
from rest_framework.serializers import as_serializer_error

from .settings import gdx_settings
from .typing import FieldError

if TYPE_CHECKING:
    from rest_framework.exceptions import ErrorDetail

    from .typing import Callable, ParamSpec, SerializerErrorData, TypeVar, ValidationErrorType

    T = TypeVar("T")
    P = ParamSpec("P")


__all__ = [
    "get_constraint_message",
    "GQLNotFoundError",
    "GQLPermissionDeniedError",
    "GQLValidationError",
]


class GQLPermissionDeniedError(GraphQLError):
    """Exception raised when the user has insufficient permissions to access a resource."""

    def __init__(self, message: str, code: str) -> None:
        super().__init__(message, extensions={"code": code})


class GQLNotFoundError(GraphQLError):
    """Exception raised when the user has insufficient permissions to access a resource."""

    def __init__(self, message: str, code: str) -> None:
        super().__init__(message, extensions={"code": code})


class GQLValidationError(GraphQLError):
    """Exception raised when mutation validation fails."""

    def __init__(self, error: ValidationErrorType) -> None:
        detail = as_serializer_error(error)
        if graphene_settings.CAMELCASE_ERRORS:
            detail = camelize(detail)
        detail = flatten_errors(detail)
        errors = to_field_errors(detail)
        super().__init__(
            gdx_settings.MUTATION_VALIDATION_ERROR_MESSAGE,
            extensions={"code": gdx_settings.MUTATION_VALIDATION_ERROR_CODE, "errors": errors},
        )


def validation_errors_to_graphql_errors(func: Callable[P, T]) -> Callable[P, T]:
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            return func(*args, **kwargs)
        except (DjangoValidationError, SerializerValidationError) as error:
            raise GQLValidationError(error) from error

    return wrapper


CONSTRAINT_PATTERNS: tuple[re.Pattern, ...] = (
    # Postgres
    re.compile(r'^new row for relation "(?P<relation>\w+)" violates check constraint "(?P<constraint>\w+)"'),
    re.compile(r'^duplicate key value violates unique constraint "(?P<constraint>\w+)"'),
    # SQLite
    re.compile(r"^CHECK constraint failed: (?P<constraint>\w+)$"),
    re.compile(r"^UNIQUE constraint failed: (?P<fields>[\w., ]+)$"),
)


def get_constraint_message(message: str) -> str:
    """Try to get the error message for a constraint violation from the model meta constraints."""
    if (match := CONSTRAINT_PATTERNS[0].match(message)) is not None:
        relation: str = match.group("relation")
        constraint: str = match.group("constraint")
        return postgres_check_constraint_message(relation, constraint, message)

    if (match := CONSTRAINT_PATTERNS[1].match(message)) is not None:
        constraint: str = match.group("constraint")
        return postgres_unique_constraint_message(constraint, message)

    if (match := CONSTRAINT_PATTERNS[2].match(message)) is not None:
        constraint: str = match.group("constraint")
        return sqlite_check_constraint_message(constraint, message)

    if (match := CONSTRAINT_PATTERNS[3].match(message)) is not None:
        fields: list[str] = match.group("fields").split(",")
        relation: str = fields[0].split(".")[0]
        fields = [field.strip().split(".")[1] for field in fields]
        return sqlite_unique_constraint_message(relation, fields, message)

    return message


def postgres_check_constraint_message(relation: str, constraint: str, default_message: str) -> str:
    for model in apps.get_models():
        if model._meta.db_table != relation:
            continue
        for constr in model._meta.constraints:
            if not isinstance(constr, models.CheckConstraint):
                continue  # pragma: no cover
            if constr.name == constraint:
                return constr.violation_error_message
    return default_message


def postgres_unique_constraint_message(constraint: str, default_message: str) -> str:
    for model in apps.get_models():
        for constr in model._meta.constraints:
            if not isinstance(constr, models.UniqueConstraint):
                continue  # pragma: no cover
            if constr.name == constraint:
                return constr.violation_error_message
    return default_message


def sqlite_check_constraint_message(constraint: str, default_message: str) -> str:
    for model in apps.get_models():
        for constr in model._meta.constraints:
            if not isinstance(constr, models.CheckConstraint):
                continue  # pragma: no cover
            if constr.name == constraint:
                return constr.violation_error_message
    return default_message


def sqlite_unique_constraint_message(relation: str, fields: list[str], default_message: str) -> str:
    for model in apps.get_models():
        if model._meta.db_table != relation:
            continue
        for constr in model._meta.constraints:
            if not isinstance(constr, models.UniqueConstraint):
                continue  # pragma: no cover
            if set(constr.fields) == set(fields):
                return constr.violation_error_message
    return default_message


def flatten_errors(errors: dict[str, list[ErrorDetail | str] | SerializerErrorData]) -> SerializerErrorData:
    """
    Flatten nested errors dict to a single level.

    >>> a = {"billing_address": {"city": ["msg1"], "post_code": ["msg2"]}}
    >>> flatten_errors(a)
    {"billing_address.city": ["msg1"], "billing_address.post_code": ["msg2"]}
    """
    flattened_errors: dict[str, list[ErrorDetail]] = {}
    for field, error in errors.items():
        if isinstance(error, dict):
            for inner_field, inner_error in flatten_errors(error).items():
                flattened_errors[f"{field}.{inner_field}"] = inner_error
        else:
            flattened_errors[field] = error

    return flattened_errors


def to_field_errors(errors: SerializerErrorData) -> list[FieldError]:
    """
    Convert a flattened errors dict to a list of field errors.

    >>> a = {
    ...     "city": [
    ...         ErrorDetail(string="msg1", code="foo"),
    ...         ErrorDetail(string="msg2", code="bar"),
    ...     ],
    ...     "post_code": ["msg3"],
    ... }
    ...
    >>> to_field_errors(a)
    [
        {"field": "city", "message": "msg1", "code": "foo"},
        {"field": "city", "message": "msg2", "code": "bar"},
        {"field": "post_code", "message": "msg3", "code": ""},
    ]
    """
    return [
        FieldError(
            field=field,
            message=message,
            code=getattr(message, "code", ""),
        )
        for field, messages in errors.items()
        for message in messages
    ]
